// Copyright (C) 2019  Wilko Manger
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'package:matrix_sdk/matrix_sdk.dart';
import 'package:matrix_sdk/src/event/room/message_event.dart';
import 'package:matrix_sdk/src/event/room/state/avatar_change_event.dart';
import 'package:matrix_sdk/src/event/room/state/member_change_event.dart';
import 'package:matrix_sdk/src/event/room/state/name_change_event.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/util/json.dart';
import 'package:meta/meta.dart';

import 'room/redaction_event.dart';

abstract class Event extends Identifiable<EventId> implements JsonWritable {
  @protected
  String get matrixType;

  @override
  final EventId id;

  final SentState sentState = SentState.sent;

  final UnsignedData unsigned;

  EventContent get content;

  Event(this.id, this.unsigned);

  /// Create event from the given [json].
  ///
  /// If [user] is specified, will use that instance as
  /// user state info.
  ///
  /// Else, if [users] is specified, will look in that map
  /// and add possible user state info to the event.
  ///
  /// Otherwise it will create a new [User] instance with just
  /// the id.
  static Future<Event> fromJson(RoomEventArgs args, Map<String, dynamic> json) async {
    final type = json['type'];
    final stateKey = json['state_key'];

    final finalArgs = await args.copyWithJson(json);

    RoomEvent event;
    switch (type) {
      case NameChangeEvent.type:
        event = NameChangeEvent.fromJson(finalArgs, json);
        break;
      case AvatarChangeEvent.type:
        event = AvatarChangeEvent.fromJson(finalArgs, json);
        break;
      case MemberChangeEvent.type:
        event =  MemberChangeEvent.fromJson(finalArgs, stateKey, json);
        break;
      case MessageEvent.type:
        event = MessageEvent.fromJson(finalArgs, json);
        break;
      case RedactionEvent.type:
        event = RedactionEvent.fromJson(finalArgs, json);
        break;
    }

    if (event != null && json.containsKey('unsigned')
     && json['unsigned']?.containsKey('redacted_because') == true) {
      final redactedBecause = json['unsigned']['redacted_because'];
      if (redactedBecause != null) {
        final redaction = RedactionEvent.fromJson(
          await finalArgs.copyWithJson(redactedBecause),
          redactedBecause
        );

        final redacted = RedactedEvent(
          redaction: redaction,
          original: event
        );

        event = redacted;
      }
    }

    return event;
  }

  factory Event.fromContent(EventContent content, RoomEventArgs args) {
    if (content is TextMessage) {
      return TextMessageEvent(content, args);
    } else {
      return null;
    }
  }

  @override
  Map<String, dynamic> toJson() => {
    'event_id': id.toJson(),
    'type': matrixType,
    'content': content?.toJson() ?? {},
    'unsigned': unsigned?.toJson()
  };

  @override
  String toString() => toJson().toString();
}

enum SentState {
  unsent,
  failed,
  sent
}

abstract class EventContent {

  String get matrixType;

  Map<String, dynamic> toJson() { return {}; }
}

class UnsignedData {
  final String transactionId;

  UnsignedData({@required this.transactionId});

  dynamic toJson() => {
    'transaction_id': transactionId
  };
}